/*
 * Implementation of hyperlinking (hlink.dll)
 *
 * Copyright 2005 Aric Stewart for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdarg.h>

#define COBJMACROS

#include "winerror.h"
#include "windef.h"
#include "winbase.h"
#include "winuser.h"
#include "ole2.h"
#include "unknwn.h"

#include "wine/debug.h"
#include "hlink.h"

#include "initguid.h"
#include "hlguids.h"

WINE_DEFAULT_DEBUG_CHANNEL(hlink);

typedef HRESULT (CALLBACK *LPFNCREATEINSTANCE)(IUnknown* pUnkOuter, REFIID riid,
         LPVOID* ppvObject);

static const IClassFactoryVtbl hlcfvt;
extern HRESULT WINAPI HLink_Constructor(IUnknown *pUnkOuter, REFIID riid, 
        LPVOID* ppv);
extern HRESULT WINAPI HLinkBrowseContext_Constructor(IUnknown *pUnkOuter, 
        REFIID riid, LPVOID* ppv);

typedef struct
{
    const IClassFactoryVtbl *lpVtbl;
    LONG                    ref;
    CLSID                   *rclsid;
    LPFNCREATEINSTANCE      lpfnCI;
    const   IID             *riidInst;
} CFImpl;

static const IHlinkVtbl hlvt;
static IClassFactory * base_fnConstructor(LPFNCREATEINSTANCE lpfnCI, REFIID riidInst);

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	TRACE("(0x%p, %ld, %p)\n",hinstDLL,fdwReason,lpvReserved);

	if (fdwReason == DLL_PROCESS_ATTACH)
	{
        DisableThreadLibraryCalls(hinstDLL);
		/* FIXME: Initialisation */
	}
	else if (fdwReason == DLL_PROCESS_DETACH)
	{
		/* FIXME: Cleanup */
	}

	return TRUE;
}

HRESULT WINAPI DllCanUnloadNow( void )
{
    FIXME("STUB\n");
    return S_OK;
}

HRESULT WINAPI HlinkCreateFromMoniker( IMoniker *pimkTrgt, LPCWSTR pwzLocation, 
        LPCWSTR pwzFriendlyName, IHlinkSite* pihlsite, DWORD dwSiteData, 
        IUnknown* piunkOuter, REFIID riid, void* ppvObj)
{
    HRESULT r = S_OK;

    TRACE("%p %s %s %p %li %p %s %p\n", pimkTrgt, debugstr_w(pwzLocation), 
            debugstr_w(pwzFriendlyName), pihlsite, dwSiteData, piunkOuter, 
            debugstr_guid(riid), ppvObj);

    r = CoCreateInstance(&CLSID_StdHlink,piunkOuter,CLSCTX_INPROC_SERVER, riid, 
            ppvObj);
    
    if (r == S_OK)
    {
        if (pwzLocation)
            IHlink_SetStringReference(*(IHlink**)ppvObj, HLINKSETF_LOCATION,
                    NULL, pwzLocation);
        if (pwzFriendlyName)
            IHlink_SetFriendlyName(*(IHlink**)ppvObj, pwzFriendlyName);
        if (pihlsite)
            IHlink_SetHlinkSite(*(IHlink**)ppvObj, pihlsite, dwSiteData);
        if (pimkTrgt)
            IHlink_SetMonikerReference(*(IHlink**)ppvObj, 0, pimkTrgt,
                    pwzLocation);
    }
    
    TRACE("Returning %li\n",r);

    return r;
}

HRESULT WINAPI HlinkCreateFromString( LPCWSTR pwzTarget, LPCWSTR pwzLocation, 
        LPCWSTR pwzFriendlyName, IHlinkSite* pihlsite, DWORD dwSiteData, 
        IUnknown* piunkOuter, REFIID riid, void* ppvObj)
{
    HRESULT r = S_OK;

    TRACE("%s %s %s %p %li %p %s %p\n", debugstr_w(pwzTarget),
            debugstr_w(pwzLocation), debugstr_w(pwzFriendlyName), pihlsite,
            dwSiteData, piunkOuter, debugstr_guid(riid), ppvObj);

    r = CoCreateInstance(&CLSID_StdHlink,piunkOuter,CLSCTX_INPROC_SERVER, riid, 
            ppvObj);
   

    if (r == S_OK)
    {
        if (pwzLocation)
            IHlink_SetStringReference(*(IHlink**)ppvObj, 
                    HLINKSETF_LOCATION, NULL, pwzLocation);

        if (pwzTarget)
        {
            IMoniker *pTgtMk = NULL;
            IBindCtx *pbc = NULL;
            ULONG eaten;

            CreateBindCtx(0, &pbc);
            r = MkParseDisplayName(pbc, pwzTarget, &eaten, &pTgtMk);
            IBindCtx_Release(pbc);

            if (r != S_OK)
            {
                FIXME("ParseDisplayName failed. Falling back to file\n");
                r = CreateFileMoniker(pwzTarget,&pTgtMk);
            }

            if (pTgtMk)
            {
                IHlink_SetMonikerReference(*(IHlink**)ppvObj, 0, pTgtMk,
                    pwzLocation);
                IMoniker_Release(pTgtMk);
            }
            else
                FIXME("Unable to come up with a Moniker, Expect Problems\n");

            IHlink_SetStringReference(*(IHlink**)ppvObj, 
                    HLINKSETF_TARGET, pwzTarget, NULL);
        }

        if (pwzFriendlyName)
            IHlink_SetFriendlyName(*(IHlink**)ppvObj, pwzFriendlyName);
        if (pihlsite)
            IHlink_SetHlinkSite(*(IHlink**)ppvObj, pihlsite, dwSiteData);
    }
    
    TRACE("Returning %li\n",r);

    return r;
}


HRESULT WINAPI HlinkCreateBrowseContext( IUnknown* piunkOuter, REFIID riid,
        void* ppvObj)
{
    HRESULT r = S_OK;

    TRACE("%p %s %p\n", piunkOuter, debugstr_guid(riid), ppvObj);

    r = CoCreateInstance(&CLSID_StdHlinkBrowseContext, piunkOuter, 
            CLSCTX_INPROC_SERVER, riid, ppvObj);
    
    TRACE("returning %li\n",r);

    return r;
}

HRESULT WINAPI HlinkNavigate(IHlink *phl, IHlinkFrame *phlFrame, 
        DWORD grfHLNF, LPBC pbc, IBindStatusCallback *pbsc, 
        IHlinkBrowseContext *phlbc)
{
    HRESULT r = S_OK;

    TRACE("%p %p %li %p %p %p\n", phl, phlFrame, grfHLNF, pbc, pbsc, phlbc);

    if (phlFrame)
        r = IHlinkFrame_Navigate(phlFrame,grfHLNF, pbc, pbsc, phl);
    else if (phl)
        r = IHlink_Navigate(phl, grfHLNF,pbc, pbsc, phlbc);

    return r;
}

HRESULT WINAPI HlinkOnNavigate( IHlinkFrame *phlFrame, 
        IHlinkBrowseContext* phlbc, DWORD grfHLNF, IMoniker *pmkTarget, 
        LPCWSTR pwzLocation, LPCWSTR pwzFriendlyName, ULONG* puHLID)
{
    HRESULT r = S_OK;

    TRACE("%p %p %li %p %s %s %p\n",phlFrame, phlbc, grfHLNF, pmkTarget,
            debugstr_w(pwzLocation), debugstr_w(pwzFriendlyName), puHLID);

    r = IHlinkBrowseContext_OnNavigateHlink(phlbc, grfHLNF, pmkTarget,
            pwzLocation, pwzFriendlyName, puHLID);

    if (phlFrame)
        r = IHlinkFrame_OnNavigate(phlFrame,grfHLNF,pmkTarget, pwzLocation, 
                pwzFriendlyName, 0);

    return r;
}

HRESULT WINAPI DllGetClassObject(REFCLSID rclsid, REFIID iid, LPVOID *ppv)
{
    HRESULT hres = E_OUTOFMEMORY;
    IClassFactory   *pcf = NULL;

    TRACE("%s %s %p\n", debugstr_guid(rclsid), debugstr_guid(iid), ppv);

    if (!ppv) return E_INVALIDARG;
    *ppv = NULL;

    if (IsEqualIID(rclsid, &CLSID_StdHlink))
        pcf = base_fnConstructor(&HLink_Constructor, NULL);
    else if (IsEqualIID(rclsid, &CLSID_StdHlinkBrowseContext))
        pcf = base_fnConstructor(&HLinkBrowseContext_Constructor, NULL);
    else
        return CLASS_E_CLASSNOTAVAILABLE;
   
    hres = IClassFactory_QueryInterface(pcf, iid, ppv);
    IClassFactory_Release(pcf);

    return hres;
}

HRESULT WINAPI HlinkCreateExtensionServices(LPCWSTR pwzAdditionalHeaders,
        HWND phwnd, LPCWSTR pszUsername, LPCWSTR pszPassword, 
        IUnknown *punkOuter, REFIID riid, void** ppvObj)
{
    FIXME("STUB (%s %p %s %s %p %s %p\n",debugstr_w(pwzAdditionalHeaders),
            phwnd, debugstr_w(pszUsername), debugstr_w(pszPassword),
            punkOuter, debugstr_guid(riid), ppvObj);
    return E_NOTIMPL;
}

static IClassFactory* base_fnConstructor(LPFNCREATEINSTANCE lpfnCI, 
        REFIID riidInst)
{
    CFImpl* lpclf;

    TRACE("\n");
    
    lpclf = HeapAlloc(GetProcessHeap(),0,sizeof(CFImpl));
    lpclf->ref = 1;
    lpclf->lpVtbl = &hlcfvt;
    lpclf->lpfnCI = lpfnCI;
    lpclf->riidInst = riidInst;

    return (LPCLASSFACTORY)lpclf;
}

static HRESULT WINAPI HLinkCF_fnQueryInterface ( LPCLASSFACTORY iface, 
        REFIID riid, LPVOID *ppvObj)
{
    CFImpl *This = (CFImpl *)iface;
    TRACE("(%p)->(%s)\n",This,debugstr_guid(riid));

    *ppvObj = NULL;

    if(IsEqualIID(riid, &IID_IUnknown) || IsEqualIID(riid, &IID_IClassFactory))
    {
        *ppvObj = This;
        InterlockedIncrement(&This->ref);
        return S_OK;
    }

    TRACE("-- E_NOINTERFACE\n");
    return E_NOINTERFACE;
}

static ULONG WINAPI HLinkCF_fnAddRef (LPCLASSFACTORY iface)
{
    CFImpl *This = (CFImpl *)iface;
    ULONG refCount = InterlockedIncrement(&This->ref);

    TRACE("(%p)->(count=%lu)\n", This, refCount - 1);

    return refCount;
}

static ULONG WINAPI HLinkCF_fnRelease(LPCLASSFACTORY iface)
{
    CFImpl *This = (CFImpl *)iface;
    ULONG refCount = InterlockedDecrement(&This->ref);

    TRACE("(%p)->(count=%lu)\n", This, refCount + 1);

    if (!refCount)
    {
        TRACE("-- destroying IClassFactory(%p)\n",This);
        HeapFree(GetProcessHeap(),0,This);
        return 0;
    }
    return refCount;
}

static HRESULT WINAPI HLinkCF_fnCreateInstance( LPCLASSFACTORY iface, 
        LPUNKNOWN pUnkOuter, REFIID riid, LPVOID *ppvObject)
{
    CFImpl *This = (CFImpl *)iface;
    TRACE("%p->(%p,%s,%p)\n",This,pUnkOuter,debugstr_guid(riid),ppvObject);

    *ppvObject = NULL;

    if ( This->riidInst==NULL || IsEqualCLSID(riid, This->riidInst) ||
         IsEqualCLSID(riid, &IID_IUnknown) )
    {
        return This->lpfnCI(pUnkOuter, riid, ppvObject);
    }

    ERR("unknown IID requested %s\n",debugstr_guid(riid));
    return E_NOINTERFACE;
}

static HRESULT WINAPI HLinkCF_fnLockServer(LPCLASSFACTORY iface, BOOL fLock)
{
    return E_NOTIMPL;
}

static const IClassFactoryVtbl hlcfvt = 
{
    HLinkCF_fnQueryInterface,
    HLinkCF_fnAddRef,
    HLinkCF_fnRelease,
    HLinkCF_fnCreateInstance,
    HLinkCF_fnLockServer
};

